import json
import logging


class Projects(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, id, **kwargs):
        resp = await self.proximl._query(f"/project/{id}", "GET", kwargs)
        return Project(self.proximl, **resp)

    async def list(self, **kwargs):
        resp = await self.proximl._query(f"/project", "GET", kwargs)
        projects = [Project(self.proximl, **project) for project in resp]
        return projects

    async def create(self, name, copy_keys=False, **kwargs):
        data = dict(
            name=name,
            copy_keys=copy_keys,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info(f"Creating Project {name}")
        resp = await self.proximl._query("/project", "POST", None, payload)
        project = Project(self.proximl, **resp)
        logging.info(f"Created Project {name} with id {project.id}")

        return project

    async def remove(self, id, **kwargs):
        await self.proximl._query(f"/project/{id}", "DELETE", kwargs)


class ProjectDatastore:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._datastore = kwargs
        self._id = self._datastore.get("id")
        self._project_uuid = self._datastore.get("project_uuid")
        self._name = self._datastore.get("name")
        self._type = self._datastore.get("type")
        self._region_uuid = self._datastore.get("region_uuid")

    @property
    def id(self) -> str:
        return self._id

    @property
    def project_uuid(self) -> str:
        return self._project_uuid

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    def __str__(self):
        return json.dumps({k: v for k, v in self._datastore.items()})

    def __repr__(self):
        return f"ProjectDatastore( proximl , **{self._datastore.__repr__()})"

    def __bool__(self):
        return bool(self._id)


class ProjectReservation:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._reservation = kwargs
        self._id = self._reservation.get("id")
        self._project_uuid = self._reservation.get("project_uuid")
        self._name = self._reservation.get("name")
        self._type = self._reservation.get("type")
        self._hostname = self._reservation.get("hostname")
        self._resource = self._reservation.get("resource")
        self._region_uuid = self._reservation.get("region_uuid")

    @property
    def id(self) -> str:
        return self._id

    @property
    def project_uuid(self) -> str:
        return self._project_uuid

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._type

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def resource(self) -> str:
        return self._resource

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    def __str__(self):
        return json.dumps({k: v for k, v in self._reservation.items()})

    def __repr__(self):
        return (
            f"ProjectReservation( proximl , **{self._reservation.__repr__()})"
        )

    def __bool__(self):
        return bool(self._id)


class Project:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._project = kwargs
        self._id = self._project.get("id")
        self._name = self._project.get("name")
        self._is_owner = self._project.get("owner")
        self._owner_name = self._project.get("owner_name")

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_owner(self) -> bool:
        return self._is_owner

    @property
    def owner_name(self) -> str:
        return self._owner_name

    def __str__(self):
        return json.dumps({k: v for k, v in self._project.items()})

    def __repr__(self):
        return f"Project( proximl , **{self._project.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(f"/project/{self._id}", "DELETE")

    async def list_datastores(self):
        resp = await self.proximl._query(
            f"/project/{self._id}/datastores", "GET"
        )
        datastores = [
            ProjectDatastore(self.proximl, **datastore) for datastore in resp
        ]
        return datastores

    async def list_reservations(self):
        resp = await self.proximl._query(
            f"/project/{self._id}/reservations", "GET"
        )
        reservations = [
            ProjectReservation(self.proximl, **reservation)
            for reservation in resp
        ]
        return reservations

    async def refresh_datastores(self):
        await self.proximl._query(f"/project/{self._id}/datastores", "PATCH")

    async def refresh_reservations(self):
        await self.proximl._query(f"/project/{self._id}/reservations", "PATCH")



import re
import sys
import asyncio
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.models]


@mark.create
@mark.asyncio
class GetModelTests:
    @fixture(scope="class")
    async def model(self, proximl):
        model = await proximl.models.create(
            name="CLI Automated",
            source_type="git",
            source_uri="git@github.com:proxiML/environment-tests.git",
        )
        model = await model.wait_for("ready", 120)
        yield model
        await model.remove()
        model = await model.wait_for("archived", 60)

    async def test_get_models(self, proximl, model):
        models = await proximl.models.list()
        assert len(models) > 0

    async def test_get_model(self, proximl, model):
        response = await proximl.models.get(model.id)
        assert response.id == model.id

    async def test_model_properties(self, model):
        assert isinstance(model.id, str)
        assert isinstance(model.status, str)
        assert isinstance(model.name, str)
        assert isinstance(model.size, int)

    async def test_model_str(self, model):
        string = str(model)
        regex = r"^{.*\"model_uuid\": \"" + model.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    async def test_model_repr(self, model):
        string = repr(model)
        regex = (
            r"^Model\( proximl , \*\*{.*'model_uuid': '"
            + model.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)


@mark.create
@mark.asyncio
async def test_model_aws(proximl, capsys):
    model = await proximl.models.create(
        name="CLI Automated AWS",
        source_type="aws",
        source_uri="s3://proximl-examples/models/mxnet-model.zip",
    )
    model = await model.wait_for("ready", 300)
    status = model.status
    size = model.size
    await model.remove()
    assert status == "ready"
    assert size >= 1000000


@mark.create
@mark.asyncio
async def test_model_local(proximl, capsys):
    model = await proximl.models.create(
        name="CLI Automated Local",
        source_type="local",
        source_uri="~/tensorflow-model",
    )
    attach_task = asyncio.create_task(model.attach())
    connect_task = asyncio.create_task(model.connect())
    await asyncio.gather(attach_task, connect_task)
    await model.disconnect()
    await model.refresh()
    status = model.status
    size = model.size
    await model.remove()
    assert status == "ready"
    assert size >= 1000000
    captured = capsys.readouterr()
    sys.stdout.write(captured.out)
    sys.stderr.write(captured.err)
    assert "Starting data upload from local" in captured.out
    assert "official/LICENSE  11456 bytes" in captured.out
    assert "Upload complete" in captured.out



import click
from proximl.cli import cli, pass_config, search_by_id_name


@cli.group()
@pass_config
def project(config):
    """proxiML project commands."""
    pass


@project.command()
@pass_config
def list(config):
    """List projects."""
    data = [
        ["ID", "NAME", "OWNER", "MINE"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    projects = config.proximl.run(config.proximl.client.projects.list())

    for project in projects:
        data.append(
            [
                project.id,
                project.name,
                project.owner_name,
                "X" if project.is_owner else "",
            ]
        )

    for row in data:
        click.echo(
            "{: >38.36} {: >30.28} {: >15.13} {: >4.4}" "".format(*row),
            file=config.stdout,
        )


@project.command()
@click.argument("name", type=click.STRING)
@pass_config
def create(config, name):
    """
    Create a project.

    Project is created with the specified NAME.
    """

    return config.proximl.run(
        config.proximl.client.projects.create(
            name=name,
        )
    )


@project.command()
@click.argument("project", type=click.STRING)
@pass_config
def remove(config, project):
    """
    Remove a project.

    PROJECT may be specified by name or ID, but ID is preferred.
    """
    projects = config.proximl.run(config.proximl.client.projects.list())

    found = search_by_id_name(project, projects)
    if None is found:
        raise click.UsageError("Cannot find specified project.")

    return config.proximl.run(found.remove())


@project.command()
@pass_config
def list_datastores(config):
    """List project datastores."""
    data = [
        ["ID", "NAME", "TYPE", "REGION_UUID"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]
    project = config.proximl.run(
        config.proximl.client.projects.get(config.proximl.client.project)
    )

    datastores = config.proximl.run(project.list_datastores())

    for datastore in datastores:
        data.append(
            [
                datastore.id,
                datastore.name,
                datastore.type,
                datastore.region_uuid,
            ]
        )

    for row in data:
        click.echo(
            "{: >38.36} {: >30.28} {: >15.13} {: >38.36}" "".format(*row),
            file=config.stdout,
        )


@project.command()
@pass_config
def list_reservations(config):
    """List project reservations."""
    data = [
        ["ID", "NAME", "TYPE", "RESOURCE", "HOSTNAME", "REGION_UUID"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]
    project = config.proximl.run(
        config.proximl.client.projects.get(config.proximl.client.project)
    )

    reservations = config.proximl.run(project.list_reservations())

    for reservation in reservations:
        data.append(
            [
                reservation.id,
                reservation.name,
                reservation.type,
                reservation.resource,
                reservation.hostname,
                reservation.region_uuid,
            ]
        )

    for row in data:
        click.echo(
            "{: >38.36} {: >30.28} {: >8.6} {: >15.13} {: >30.28} {: >38.36}"
            "".format(*row),
            file=config.stdout,
        )



import json
import logging
from datetime import datetime


class Providers(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, id):
        resp = await self.proximl._query(f"/provider/{id}", "GET")
        return Provider(self.proximl, **resp)

    async def list(self):
        resp = await self.proximl._query(f"/provider", "GET")
        providers = [Provider(self.proximl, **provider) for provider in resp]
        return providers

    async def enable(self, type, **kwargs):
        data = dict(type=type, **kwargs)
        payload = {k: v for k, v in data.items() if v is not None}
        logging.info(f"Enabling Provider {type}")
        resp = await self.proximl._query("/provider", "POST", None, payload)
        provider = Provider(self.proximl, **resp)
        logging.info(f"Enabled Provider {type} with id {provider.id}")

        return provider

    async def remove(self, id):
        await self.proximl._query(f"/provider/{id}", "DELETE")


class Provider:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._provider = kwargs
        self._id = self._provider.get("provider_uuid")
        self._type = self._provider.get("type")
        self._credits = self._provider.get("credits")

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> str:
        return self._type

    @property
    def credits(self) -> float:
        return self._credits

    def __str__(self):
        return json.dumps({k: v for k, v in self._provider.items()})

    def __repr__(self):
        return f"Provider( proximl , **{self._provider.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(f"/provider/{self._id}", "DELETE")

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self



import re
from pytest import mark, fixture

pytestmark = [mark.sdk, mark.integration, mark.environments]


@fixture(scope="module")
async def environments(proximl):
    environments = await proximl.environments.list()
    yield environments


@fixture(scope="module")
async def environment(environments):
    environment = next(
        (environment for environment in environments if environment.version),
        None,
    )
    yield environment


@mark.asyncio
async def test_get_environments(environments):
    assert len(environments) > 0


@mark.asyncio
async def test_environment_properties(environment):
    assert isinstance(environment.id, str)
    assert isinstance(environment.name, str)
    assert isinstance(environment.py_version, str)
    assert isinstance(environment.framework, str)
    assert isinstance(environment.version, str)
    assert isinstance(environment.cuda_version, str)


@mark.asyncio
async def test_environment_str(environment):
    string = str(environment)
    regex = r"^{.*\"id\": \"" + environment.id + r"\".*}$"
    assert isinstance(string, str)
    assert re.match(regex, string)


@mark.asyncio
async def test_environment_repr(environment):
    string = repr(environment)
    regex = (
        r"^Environment\( proximl , \*\*{.*'id': '"
        + environment.id
        + r"'.*}\)$"
    )
    assert isinstance(string, str)
    assert re.match(regex, string)



import json
import os
import shutil
import asyncio
import aiohttp
import aiodocker
import zipfile
import re
import logging

from .exceptions import (
    ConnectionError,
    ApiError,
    SpecificationError,
    ProxiMLException,
)
from aiodocker.exceptions import DockerError


VPN_IMAGE = "proximl/connection:no-upnp"
STORAGE_IMAGE = "proximl/local-storage"
STATUSES = dict(
    UNKNOWN="unknown",
    NEW="new",
    CONNECTING="connecting",
    CONNECTED="connected",
    NOT_CONNECTED="not connected",
    STOPPED="stopped",
    REMOVED="removed",
)
CONFIG_DIR = os.path.expanduser(
    os.environ.get("PROXIML_CONFIG_DIR") or "~/.proximl"
)


class Connections(object):
    def __init__(self, proximl):
        self.proximl = proximl
        self.dir = f"{CONFIG_DIR}/connections/{self.proximl.project}"
        os.makedirs(
            self.dir,
            exist_ok=True,
        )

    async def list(self):
        con_dirs = os.listdir(self.dir)
        connections = []
        con_tasks = []
        for con_dir in con_dirs:
            try:
                con_type, con_id = con_dir.split("_")
            except ValueError:
                # unintelligible directory
                continue
            connection = Connection(self.proximl, con_type, con_id)
            connections.append(connection)
            con_task = asyncio.create_task(connection.check())
            con_tasks.append(con_task)
        await asyncio.gather(*con_tasks)
        return connections

    async def cleanup_connections(self):
        logging.info("Start cleanup connections")
        con_dirs = os.listdir(self.dir)
        con_tasks = []
        logging.debug(f"con_dirs: {con_dirs}")
        for con_dir in con_dirs:
            try:
                con_type, con_id = con_dir.split("_")
            except ValueError:
                # unintelligible directory
                logging.debug(f"unintelligible con_dir: {con_dir}")
                shutil.rmtree(con_dir)
                continue
            connection = Connection(self.proximl, con_type, con_id)
            con_task = asyncio.create_task(connection._validate_entity())
            con_tasks.append(con_task)
        await asyncio.gather(*con_tasks, return_exceptions=True)
        await self.cleanup_containers(project=self.proximl.project)
        logging.info("Finish cleanup connections")

    async def cleanup_containers(self, project=None):
        logging.info("Start cleanup containers")
        if project:
            con_dirs = os.listdir(f"{CONFIG_DIR}/connections/{project}")
            await asyncio.gather(
                asyncio.create_task(
                    _cleanup_containers(
                        project,
                        self.dir,
                        con_dirs,
                        "vpn",
                    )
                ),
                asyncio.create_task(
                    _cleanup_containers(
                        project,
                        self.dir,
                        con_dirs,
                        "storage",
                    )
                ),
            )
        else:
            con_dirs = []
            proj_dirs = os.listdir(f"{CONFIG_DIR}/connections")
            logging.info(f"proj_dirs {proj_dirs}")
            for proj_dir in proj_dirs:
                con_dirs += [
                    f"{proj_dir}/{con_dir}"
                    for con_dir in os.listdir(
                        f"{CONFIG_DIR}/connections/{project}"
                    )
                ]

            logging.info(f"con_dirs {con_dirs}")
            await asyncio.gather(
                asyncio.create_task(
                    _cleanup_containers(
                        None,
                        f"{CONFIG_DIR}/connections",
                        con_dirs,
                        "vpn",
                    )
                ),
                asyncio.create_task(
                    _cleanup_containers(
                        None,
                        f"{CONFIG_DIR}/connections",
                        con_dirs,
                        "storage",
                    )
                ),
            )

        logging.info("Finish cleanup containers")

    async def remove_all(self, all_projects=False):
        if all_projects:
            proj_dirs = os.listdir(f"{CONFIG_DIR}/connections")
            for proj_dir in proj_dirs:
                shutil.rmtree(f"{CONFIG_DIR}/connections/{proj_dir}")
                os.makedirs(
                    f"{CONFIG_DIR}/connections/{proj_dir}",
                    exist_ok=True,
                )
            await self.cleanup_containers()
        else:
            shutil.rmtree(self.dir)
            os.makedirs(
                self.dir,
                exist_ok=True,
            )
            await self.cleanup_containers(project=self.proximl.project)


class Connection:
    def __init__(self, proximl, entity_type, id, entity=None, **kwargs):
        self.proximl = proximl
        self._id = id
        self._type = entity_type
        self._status = STATUSES.get("UNKNOWN")
        self._entity = entity
        CONNECTIONS_DIR = f"{CONFIG_DIR}/connections/{self.proximl.project}"
        self._dir = f"{CONNECTIONS_DIR}/{entity_type}_{id}"
        os.makedirs(
            self._dir,
            exist_ok=True,
        )

    @property
    def id(self) -> str:
        return self._id

    @property
    def type(self) -> str:
        return self._type

    @property
    def status(self) -> str:
        return self._status

    def __str__(self):
        return f"Connection for {self.type} - {self.id}: {self.status}"

    def __repr__(self):
        return f"Connection( proximl , {self.id}, {self.type})"

    async def _get_entity(self):
        if self.type == "dataset":
            self._entity = await self.proximl.datasets.get(self.id)
        elif self.type == "job":
            self._entity = await self.proximl.jobs.get(self.id)
        elif self.type == "model":
            self._entity = await self.proximl.models.get(self.id)
        elif self.type == "checkpoint":
            self._entity = await self.proximl.checkpoints.get(self.id)
        else:
            raise TypeError(
                "Connection type must be in: ['dataset', 'model', 'checkpoint', 'job']"
            )

    async def _download_connection_details(self):
        zip_file = f"{self._dir}/details.zip"
        url = await self._entity.get_connection_utility_url()
        async with aiohttp.ClientSession() as session:
            async with session.request("GET", url) as resp:
                with open(
                    zip_file,
                    "wb",
                ) as fd:
                    content = await resp.read()
                    fd.write(content)
        with zipfile.ZipFile(zip_file, "r") as zipf:
            for info in zipf.infolist():
                extracted_path = zipf.extract(info, self._dir)
                if info.create_system == 3 and os.path.isfile(
                    extracted_path
                ):  ## 3 - ZIP_UNIX_SYSTEM
                    unix_attributes = info.external_attr >> 16
                    if unix_attributes:
                        os.chmod(extracted_path, unix_attributes)

        os.remove(zip_file)

    async def _test_connection(self, container):
        entity_details = self._entity.get_connection_details()
        if not entity_details:
            return False
        net = _parse_cidr(entity_details.get("cidr"))
        target_ip = f"{net.get('first_octet')}.{net.get('second_octet')}.{net.get('third_octet')}.254"

        logging.debug("Testing connection")
        ping = await container.exec(
            ["ping", "-c", "1", target_ip],
            stdout=True,
            stderr=True,
        )
        stream = ping.start()
        await stream.read_out()
        data = await ping.inspect()
        while data["ExitCode"] is None:
            await stream.read_out()
            data = await ping.inspect()
        await stream.close()
        if data["ExitCode"] == 0:
            return True
        return False

    async def _validate_entity(self):
        try:
            await self._get_entity()
            logging.debug(f"entity: {self._entity}")
            if self._entity.status in [
                "failed",
                "finished",
                "canceled",
                "archived",
                "removed",
                "removing",
                "ready",
            ]:
                shutil.rmtree(self._dir)
                logging.debug(f"remove: {self._dir}")
                return False
            else:
                return True
        except ApiError as e:
            if e.status == 404:
                shutil.rmtree(self._dir)
                logging.debug(f"remove: {self._dir}")
                return False
            else:
                raise e

    async def check(self):
        if not self._entity:
            valid = await self._validate_entity()
            if not valid:
                self._status = STATUSES.get("REMOVED")
                return
        if not os.path.isdir(f"{self._dir}/data"):
            self._status = STATUSES.get("NEW")
            return

        try:
            with open(f"{self._dir}/vpn_id", "r") as f:
                vpn_id = f.read()
        except OSError as e:
            self._status = STATUSES.get("STOPPED")
            return

        docker = aiodocker.Docker()
        await _check_docker_status(docker)

        try:
            container = await docker.containers.get(vpn_id)
        except DockerError as e:
            if e.status == 404:
                self._status = STATUSES.get("STOPPED")
                await docker.close()
                return
            raise e

        data = await container.show()
        if not data["State"]["Running"]:
            self._status = STATUSES.get("STOPPED")
            await container.delete()
            os.remove(f"{self._dir}/vpn_id")
            try:
                with open(f"{self._dir}/storage_id", "r") as f:
                    storage_id = f.read()
                try:
                    storage_container = await docker.containers.get(storage_id)
                    await storage_container.delete(force=True)
                except DockerError as e:
                    if e.status != 404:
                        raise e
            except OSError as e:
                pass
            await docker.close()
            return

        connected = await self._test_connection(container)
        await docker.close()
        if connected:
            self._status = STATUSES.get("CONNECTED")
        else:
            self._status = STATUSES.get("NOT_CONNECTED")

    async def start(self):
        logging.info(f"Beginning start {self.type} connection {self.id}")

        if self.status == STATUSES.get("UNKNOWN"):
            await self.check()
        if self.status in [
            STATUSES.get("CONNECTING"),
            STATUSES.get("CONNECTED"),
            STATUSES.get("NOT_CONNECTED"),
        ]:
            raise SpecificationError(
                "status", "Only inactive connections can be started."
            )
        self._status = STATUSES.get("CONNECTING")
        logging.info(f"Connecting...")
        if not self._entity:
            await self._get_entity()
        if not os.path.isdir(f"{self._dir}/data"):
            await self._download_connection_details()

        docker = aiodocker.Docker()
        await _check_docker_status(docker)
        try:
            await asyncio.gather(
                docker.pull(VPN_IMAGE), docker.pull(STORAGE_IMAGE)
            )
        except DockerError as e:
            exists = await asyncio.gather(
                _image_exists(docker, VPN_IMAGE),
                _image_exists(docker, STORAGE_IMAGE),
            )
            if any([not i for i in exists]):
                raise e

        entity_details = self._entity.get_connection_details()
        if (
            entity_details.get("model_path")
            or entity_details.get("input_path")
            or entity_details.get("output_path")
        ):
            logging.debug(f"Starting storage container")
            storage_container = await docker.containers.run(
                _get_storage_container_config(
                    self.id,
                    entity_details.get("project_uuid"),
                    entity_details.get("entity_type"),
                    entity_details.get("cidr"),
                    f"{self._dir}/data",
                    entity_details.get("ssh_port"),
                    model_path=entity_details.get("model_path"),
                    input_path=entity_details.get("input_path"),
                    output_path=entity_details.get("output_path"),
                )
            )
            logging.debug(
                f"Storage container started, id: {storage_container.id}"
            )
            with open(f"{self._dir}/storage_id", "w") as f:
                f.write(storage_container.id)

        logging.debug(f"Starting VPN container")
        vpn_container = await docker.containers.run(
            _get_vpn_container_config(
                self.id,
                entity_details.get("project_uuid"),
                entity_details.get("entity_type"),
                entity_details.get("cidr"),
                f"{self._dir}/data",
            )
        )
        logging.debug(f"VPN container started, id: {vpn_container.id}")

        with open(f"{self._dir}/vpn_id", "w") as f:
            f.write(vpn_container.id)

        # cleanup_task = asyncio.create_task(
        #     self.proximl.connections.cleanup_connections()
        # )

        count = 0
        while count <= 30:
            logging.debug(f"Test connectivity attempt {count+1}")
            res = await self._test_connection(vpn_container)
            if res:
                logging.debug(f"Test connectivity successful {count+1}")
                break
            count += 1
        await docker.close()
        if count > 30:
            self._status = STATUSES.get("NOT_CONNECTED")
            raise ConnectionError(f"Unable to connect {self.type} {self.id}")
        self._status = STATUSES.get("CONNECTED")
        logging.info(f"Connection Successful.")
        # await cleanup_task
        logging.debug(f"Completed start {self.type} connection {self.id}")

    async def stop(self):
        logging.debug(f"Beginning stop {self.type} connection {self.id}")
        if not self._entity:
            await self._get_entity()
        docker = aiodocker.Docker()
        await _check_docker_status(docker)

        tasks = []
        logging.info("Disconnecting...")
        try:
            with open(f"{self._dir}/vpn_id", "r") as f:
                vpn_id = f.read()
            logging.debug(f"vpn container id: {vpn_id}")
            vpn_container = await docker.containers.get(vpn_id)
            vpn_delete_task = asyncio.create_task(
                vpn_container.delete(force=True)
            )
            tasks.append(vpn_delete_task)
            os.remove(f"{self._dir}/vpn_id")
        except OSError:
            logging.debug("vpn container not found")

        storage_delete_task = None
        try:
            with open(f"{self._dir}/storage_id", "r") as f:
                storage_id = f.read()
            logging.debug(f"storage container id: {vpn_id}")
            storage_container = await docker.containers.get(storage_id)
            storage_delete_task = asyncio.create_task(
                storage_container.delete(force=True)
            )
            tasks.append(storage_delete_task)
            os.remove(f"{self._dir}/storage_id")
        except OSError:
            logging.debug("storage container not found")

        await asyncio.gather(*tasks)
        await docker.close()
        self._status = STATUSES.get("REMOVED")
        shutil.rmtree(self._dir)
        logging.info("Disconnected.")
        logging.debug(f"Completed stop {self.type} connection {self.id}")


async def _cleanup_containers(project, path, con_dirs, type):
    containers_target = []
    for con_dir in con_dirs:
        try:
            with open(f"{path}/{con_dir}/{type}_id", "r") as f:
                id = f.read()
            containers_target.append(id)
        except OSError:
            continue

    docker = aiodocker.Docker()
    await _check_docker_status(docker)

    if project:
        filter = dict(
            label=[
                "service=proximl",
                f"type={type}",
                f"project={project}",
            ]
        )
    else:
        filter = dict(
            label=[
                "service=proximl",
                f"type={type}",
            ]
        )

    containers = await docker.containers.list(
        all=True,
        filters=json.dumps(filter),
    )

    tasks = [
        asyncio.create_task(container.delete(force=True))
        for container in containers
        if container.id not in containers_target
    ]

    await asyncio.gather(*tasks, return_exceptions=True)
    await docker.close()


def _parse_cidr(cidr):
    res = re.match(
        r"(?P<first_octet>[0-9]{1,3})\.(?P<second_octet>[0-9]{1,3})\.(?P<third_octet>[0-9]{1,3})\.(?P<fourth_octet>[0-9]{1,3})/(?P<mask_length>[0-9]{1,2})",
        cidr,
    )
    net = res.groupdict()
    return net


def _get_vpn_container_config(id, project_uuid, entity_type, cidr, data_dir):
    config = dict(
        Image=VPN_IMAGE,
        Hostname=id,
        Cmd=[],
        AttachStdin=False,
        AttachStdout=False,
        AttachStderr=False,
        Tty=False,
        Env=[
            f"NETWORK={id}",
            "DEBUG=1",
        ],
        HostConfig=dict(
            Init=True,
            Binds=[f"{data_dir}:/etc/tinc:rw"],
            NetworkMode="host",
            CapAdd=["NET_ADMIN"],
        ),
        Labels=dict(
            type="vpn",
            service="proximl",
            id=id,
            project=project_uuid,
            entity_type=entity_type,
        ),
    )
    return config


def _get_storage_container_config(
    id,
    project_uuid,
    entity_type,
    cidr,
    data_dir,
    ssh_port,
    model_path=None,
    input_path=None,
    output_path=None,
):
    Binds = [f"{data_dir}/.ssh:/opt/ssh"]
    if model_path:
        Binds.append(f"{os.path.expanduser(model_path)}:/opt/model:ro")
    if input_path:
        Binds.append(f"{os.path.expanduser(input_path)}:/opt/data:ro")
    if output_path:
        Binds.append(f"{os.path.expanduser(output_path)}:/opt/output:rw")
    config = dict(
        Image=STORAGE_IMAGE,
        Hostname=id,
        Cmd=[],
        AttachStdin=False,
        AttachStdout=False,
        AttachStderr=False,
        Tty=False,
        Env=[
            f"VPN_CIDR={cidr}",
        ],
        ExposedPorts={f"22/tcp": {}},
        HostConfig=dict(
            Init=True,
            Binds=Binds,
            PortBindings={
                f"22/tcp": [dict(HostPort=f"{ssh_port}", HostIP="0.0.0.0")],
            },
        ),
        Labels=dict(
            type="storage",
            service="proximl",
            id=id,
            project=project_uuid,
            entity_type=entity_type,
        ),
    )
    return config


async def _image_exists(client, id):
    if not id:
        return False
    try:
        await client.images.inspect(id)
        return True
    except DockerError:
        return False


async def _check_docker_status(client):
    try:
        await client.version()
    except DockerError:
        raise ProxiMLException(
            "Docker is not installed, not running, or your user does not have docker privileges. "
            + "Ensure docker is installed, then add your user to the docker group. "
            + "Commonly, this is accomplished by entering 'sudo gpasswd -a <username> docker'. "
            + "You may need to logout and log back in to your computer for this to take effect"
        )



import asyncio
from proximl.projects import Projects
from pytest import fixture, mark
from unittest.mock import Mock, AsyncMock, patch, create_autospec

from proximl.proximl import ProxiML
from proximl.auth import Auth
from proximl.datasets import Dataset, Datasets
from proximl.checkpoints import Checkpoint, Checkpoints
from proximl.models import Model, Models
from proximl.gpu_types import GpuType, GpuTypes
from proximl.environments import Environment, Environments
from proximl.jobs import Job, Jobs
from proximl.connections import Connections
from proximl.projects import (
    Projects,
    Project,
    ProjectDatastore,
    ProjectReservation,
)
from proximl.cloudbender import Cloudbender
from proximl.cloudbender.providers import Provider, Providers
from proximl.cloudbender.regions import Region, Regions
from proximl.cloudbender.nodes import Node, Nodes
from proximl.cloudbender.devices import Device, Devices
from proximl.cloudbender.datastores import Datastore, Datastores
from proximl.cloudbender.reservations import Reservation, Reservations
from proximl.cloudbender.device_configs import DeviceConfig, DeviceConfigs


pytestmark = mark.unit


@fixture(scope="session")
def mock_my_datasets():
    proximl = Mock()
    yield [
        Dataset(
            proximl,
            dataset_uuid="1",
            project_uuid="proj-id-1",
            name="first one",
            status="ready",
            size=100000000,
            createdAt="2020-12-31T23:59:59.000Z",
        ),
        Dataset(
            proximl,
            dataset_uuid="2",
            project_uuid="proj-id-1",
            name="second one",
            status="ready",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Dataset(
            proximl,
            dataset_uuid="3",
            project_uuid="proj-id-1",
            name="first one",
            status="ready",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Dataset(
            proximl,
            dataset_uuid="4",
            project_uuid="proj-id-1",
            name="other one",
            status="ready",
            size=100000000,
            createdAt="2020-12-31T23:59:59.000Z",
        ),
        Dataset(
            proximl,
            dataset_uuid="5",
            project_uuid="proj-id-1",
            name="not ready",
            status="new",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Dataset(
            proximl,
            dataset_uuid="6",
            project_uuid="proj-id-1",
            name="failed",
            status="failed",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
    ]


@fixture(scope="session")
def mock_public_datasets():
    proximl = Mock()
    yield [
        Dataset(
            proximl,
            dataset_uuid="11",
            name="first one",
            status="ready",
        ),
        Dataset(
            proximl,
            dataset_uuid="12",
            name="second one",
            status="ready",
        ),
        Dataset(
            proximl,
            dataset_uuid="15",
            name="not ready",
            status="new",
        ),
        Dataset(
            proximl,
            dataset_uuid="16",
            name="failed",
            status="failed",
        ),
    ]


@fixture(scope="session")
def mock_my_checkpoints():
    proximl = Mock()
    yield [
        Checkpoint(
            proximl,
            checkpoint_uuid="1",
            project_uuid="proj-id-1",
            name="first one",
            status="ready",
            size=100000000,
            createdAt="2020-12-31T23:59:59.000Z",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="2",
            project_uuid="proj-id-1",
            name="second one",
            status="ready",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="3",
            project_uuid="proj-id-1",
            name="first one",
            status="ready",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="4",
            project_uuid="proj-id-1",
            name="other one",
            status="ready",
            size=100000000,
            createdAt="2020-12-31T23:59:59.000Z",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="5",
            project_uuid="proj-id-1",
            name="not ready",
            status="new",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="6",
            project_uuid="proj-id-1",
            name="failed",
            status="failed",
            size=100000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
    ]


@fixture(scope="session")
def mock_public_checkpoints():
    proximl = Mock()
    yield [
        Checkpoint(
            proximl,
            checkpoint_uuid="11",
            name="first one",
            status="ready",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="12",
            name="second one",
            status="ready",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="15",
            name="not ready",
            status="new",
        ),
        Checkpoint(
            proximl,
            checkpoint_uuid="16",
            name="failed",
            status="failed",
        ),
    ]


@fixture(scope="session")
def mock_models():
    proximl = Mock()
    yield [
        Model(
            proximl,
            model_uuid="1",
            project_uuid="proj-id-1",
            name="first one",
            status="ready",
            size=10000000,
            createdAt="2020-12-31T23:59:59.000Z",
        ),
        Model(
            proximl,
            model_uuid="2",
            project_uuid="proj-id-1",
            name="second one",
            status="ready",
            size=10000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Model(
            proximl,
            model_uuid="5",
            project_uuid="proj-id-1",
            name="not ready",
            status="new",
            size=10000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
        Model(
            proximl,
            model_uuid="6",
            project_uuid="proj-id-1",
            name="failed",
            status="failed",
            size=10000000,
            createdAt="2021-01-01T00:00:01.000Z",
        ),
    ]


@fixture(scope="session")
def mock_gpu_types():
    proximl = Mock()
    yield [
        GpuType(
            proximl,
            **{
                "name": "A100",
                "price": {"min": 2.78, "max": 4.9},
                "id": "a100-id",
                "abbrv": "a100",
            },
        ),
        GpuType(
            proximl,
            **{
                "price": {"min": 1, "max": 3.94},
                "name": "V100",
                "id": "v100-id",
                "abbrv": "v100",
            },
        ),
        GpuType(
            proximl,
            **{
                "price": {"min": 0.35, "max": 0.35},
                "name": "RTX 2080 Ti",
                "id": "2080ti-id",
                "abbrv": "rtx2080ti",
            },
        ),
        GpuType(
            proximl,
            **{
                "price": {"min": 0.1, "max": 0.1},
                "name": "GTX 1060",
                "id": "1060-id",
                "abbrv": "gtx1060",
            },
        ),
    ]


@fixture(scope="function")
def mock_environments():
    proximl = Mock()
    yield [
        Environment(
            proximl,
            **{
                "id": "DEEPLEARNING_PY39",
                "framework": "Deep Learning",
                "py_version": "3.9",
                "cuda_version": "11.7",
                "name": "Deep Learning - Python 3.9",
            },
        ),
        Environment(
            proximl,
            **{
                "id": "DEEPLEARNING_PY310",
                "framework": "Deep Learning",
                "py_version": "3.10",
                "cuda_version": "12.1",
                "name": "Deep Learning - Python 3.10",
            },
        ),
        Environment(
            proximl,
            **{
                "id": "PYTORCH_PY38_17",
                "framework": "PyTorch",
                "py_version": "3.8",
                "version": "1.7",
                "cuda_version": "11.1",
                "name": "PyTorch 1.7 - Python 3.8",
            },
        ),
        Environment(
            proximl,
            **{
                "id": "PYTORCH_PY37_17",
                "framework": "PyTorch",
                "py_version": "3.7",
                "version": "1.7",
                "cuda_version": "10.2",
                "name": "PyTorch 1.7 - Python 3.7",
            },
        ),
        Environment(
            proximl,
            **{
                "id": "PYTORCH_PY37_16",
                "framework": "PyTorch",
                "py_version": "3.7",
                "version": "1.6",
                "cuda_version": "10.2",
                "name": "PyTorch 1.6 - Python 3.7",
            },
        ),
        Environment(
            proximl,
            **{
                "id": "PYTORCH_PY37_15",
                "framework": "PyTorch",
                "py_version": "3.7",
                "version": "1.5",
                "cuda_version": "10.1",
                "name": "PyTorch 1.5 - Python 3.7",
            },
        ),
        Environment(
            proximl,
            **{
                "id": "TENSORFLOW_PY38_24",
                "framework": "Tensorflow",
                "py_version": "3.8",
                "version": "2.4",
                "cuda_version": "11.1",
                "name": "Tensorflow 2.4 - Python 3.8",
            },
        ),
        Environment(
            proximl,
            **{
                "id": "TENSORFLOW_PY37_114",
                "framework": "Tensorflow",
                "py_version": "3.7",
                "version": "1.14",
                "cuda_version": "10.1",
                "name": "Tensorflow 1.14 - Python 3.7",
            },
        ),
    ]


@fixture(scope="session")
def mock_jobs():
    proximl = Mock()
    yield [
        Job(
            proximl,
            **{
                "project_uuid": "proj-id-1",
                "job_uuid": "job-id-1",
                "name": "test notebook",
                "start": "2021-02-11T15:46:22.455Z",
                "type": "notebook",
                "status": "new",
                "credits_per_hour": 0.1,
                "credits": 0.1007,
                "workers": [
                    {
                        "rig_uuid": "rig-id-1",
                        "job_worker_uuid": "worker-id-1",
                        "command": "jupyter lab",
                        "status": "new",
                    }
                ],
                "worker_status": "new",
                "resources": {
                    "gpu_count": 1,
                    "gpu_type_id": "1060-id",
                    "disk_size": 10,
                },
                "model": {
                    "size": 7176192,
                    "source_type": "git",
                    "source_uri": "git@github.com:proxiML/test-private.git",
                    "status": "new",
                },
                "data": {
                    "datasets": [
                        {
                            "dataset_uuid": "data-id-1",
                            "name": "first one",
                            "type": "public",
                            "size": 184549376,
                        },
                        {
                            "dataset_uuid": "data-id-2",
                            "name": "second one",
                            "type": "public",
                            "size": 5068061409,
                        },
                    ],
                    "status": "ready",
                },
                "environment": {
                    "type": "DEEPLEARNING_PY310",
                    "image_size": 44966989795,
                    "env": [
                        {"value": "env1val", "key": "env1"},
                        {"value": "env2val", "key": "env2"},
                    ],
                    "worker_key_types": ["aws", "gcp"],
                    "status": "new",
                },
                "vpn": {
                    "status": "new",
                    "cidr": "10.106.171.0/24",
                    "client": {
                        "port": "36017",
                        "id": "cus-id-1",
                        "address": "10.106.171.253",
                    },
                    "net_prefix_type_id": 1,
                },
                "nb_token": "token",
            },
        ),
        Job(
            proximl,
            **{
                "project_uuid": "proj-id-1",
                "job_uuid": "job-id-2",
                "name": "test training",
                "start": "2021-02-11T15:48:39.476Z",
                "stop": "2021-02-11T15:50:16.554Z",
                "type": "training",
                "status": "finished",
                "credits_per_hour": 0.3,
                "credits": 0.0054,
                "workers": [
                    {
                        "rig_uuid": "rig-id-1",
                        "job_worker_uuid": "worker-id-11",
                        "command": "PYTHONPATH=$PYTHONPATH:$ML_MODEL_PATH python -m official.vision.image_classification.resnet_cifar_main --num_gpus=1 --data_dir=$ML_DATA_PATH --model_dir=$ML_OUTPUT_PATH --enable_checkpoint_and_export=True --train_epochs=1 --batch_size=1024",
                        "status": "stopped",
                    },
                    {
                        "rig_uuid": "rig-id-2",
                        "job_worker_uuid": "worker-id-12",
                        "command": "PYTHONPATH=$PYTHONPATH:$ML_MODEL_PATH python -m official.vision.image_classification.resnet_cifar_main --num_gpus=1 --data_dir=$ML_DATA_PATH --model_dir=$ML_OUTPUT_PATH --enable_checkpoint_and_export=True --train_epochs=1 --batch_size=1024",
                        "status": "stopped",
                    },
                    {
                        "rig_uuid": "rig-id-2",
                        "job_worker_uuid": "worker-id-13",
                        "command": "PYTHONPATH=$PYTHONPATH:$ML_MODEL_PATH python -m official.vision.image_classification.resnet_cifar_main --num_gpus=1 --data_dir=$ML_DATA_PATH --model_dir=$ML_OUTPUT_PATH --enable_checkpoint_and_export=True --train_epochs=1 --batch_size=1024",
                        "status": "stopped",
                    },
                ],
                "worker_status": "stopped",
                "resources": {
                    "gpu_count": 1,
                    "gpu_type_id": "1060-id",
                    "disk_size": 10,
                },
                "model": {
                    "size": 7086080,
                    "source_type": "git",
                    "source_uri": "git@github.com:proxiML/test-private.git",
                    "status": "ready",
                },
                "data": {
                    "datasets": [
                        {
                            "dataset_uuid": "data-id-1",
                            "name": "first one",
                            "type": "public",
                            "size": 184549376,
                        }
                    ],
                    "status": "ready",
                    "output_type": "aws",
                    "output_uri": "s3://proximl-examples/output/resnet_cifar10",
                },
                "environment": {
                    "type": "DEEPLEARNING_PY37",
                    "image_size": 39656398629,
                    "env": [],
                    "worker_key_types": [],
                    "status": "ready",
                },
                "vpn": {
                    "status": "stopped",
                    "cidr": "10.222.241.0/24",
                    "client": {
                        "port": "36978",
                        "id": "cus-id-1",
                        "address": "10.222.241.253",
                    },
                    "net_prefix_type_id": 1,
                },
            },
        ),
    ]


@fixture(scope="session")
def mock_projects():
    proximl = Mock()
    yield [
        Project(
            proximl,
            **{
                "id": "proj-id-1",
                "name": "Personal",
                "owner": True,
                "owner_name": "Me",
                "created_name": "Me",
                "job_all": True,
                "dataset_all": True,
                "model_all": True,
                "createdAt": "2020-12-31T23:59:59.000Z",
            },
        ),
        Project(
            proximl,
            **{
                "id": "project-id-2",
                "name": "My Other Project",
                "owner": True,
                "owner_name": "Me",
                "created_name": "Me",
                "job_all": True,
                "dataset_all": True,
                "model_all": True,
                "createdAt": "2020-12-31T23:59:59.000Z",
            },
        ),
        Project(
            proximl,
            **{
                "id": "project-id-3",
                "name": "Someone Elses Project",
                "owner": False,
                "owner_name": "Someone Else",
                "created_name": "Someone Else",
                "job_all": False,
                "dataset_all": False,
                "model_all": False,
                "createdAt": "2020-12-31T23:59:59.000Z",
            },
        ),
    ]


@fixture(scope="session")
def mock_providers():
    proximl = Mock()
    yield [
        Provider(
            proximl,
            **{
                "customer_uuid": "cust-id-1",
                "provider_uuid": "prov-id-1",
                "type": "physical",
                "credits": 10.0,
                "payment_mode": "stripe",
            },
        ),
        Provider(
            proximl,
            **{
                "customer_uuid": "cust-id-1",
                "provider_uuid": "prov-id-2",
                "type": "gcp",
                "credits": 0.0,
                "payment_mode": "credits",
                "credentials": {
                    "project": "proj-1",
                    "id": "gcp@serviceaccount.com",
                },
            },
        ),
    ]


@fixture(scope="session")
def mock_regions():
    proximl = Mock()
    yield [
        Region(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "provider_type": "physical",
                "name": "Physical Region 1",
                "status": "healthy",
            },
        ),
        Region(
            proximl,
            **{
                "provider_uuid": "prov-id-2",
                "region_uuid": "reg-id-2",
                "provider_type": "gcp",
                "name": "Cloud Region 1",
                "status": "healthy",
            },
        ),
    ]


@fixture(scope="session")
def mock_nodes():
    proximl = Mock()
    yield [
        Node(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "rig_uuid": "rig-id-1",
                "type": "permanent",
                "service": "compute",
                "friendly_name": "hq-a100-01",
                "hostname": "hq-a100-01",
                "status": "active",
                "online": True,
                "maintenance_mode": False,
            },
        ),
        Node(
            proximl,
            **{
                "provider_uuid": "prov-id-2",
                "region_uuid": "reg-id-2",
                "rig_uuid": "rig-id-2",
                "type": "ephemeral",
                "service": "compute",
                "friendly_name": "gcp-a100-01",
                "hostname": "gcp-a100-01",
                "status": "active",
                "online": True,
                "maintenance_mode": False,
            },
        ),
        Node(
            proximl,
            **{
                "provider_uuid": "prov-id-2",
                "region_uuid": "reg-id-2",
                "rig_uuid": "rig-id-3",
                "type": "permanent",
                "service": "storage",
                "friendly_name": "gcp-storage-01",
                "hostname": "gcp-storage-01",
                "status": "active",
                "online": True,
                "maintenance_mode": False,
            },
        ),
    ]


@fixture(scope="session")
def mock_devices():
    proximl = Mock()
    yield [
        Device(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "device_id": "dev-id-1",
                "type": "device",
                "service": "compute",
                "friendly_name": "hq-orin-01",
                "hostname": "hq-orin-01",
                "status": "active",
                "online": True,
                "maintenance_mode": False,
                "job_status": "running",
                "job_last_deployed": "2023-06-02T21:22:40.084Z",
                "job_config_id": "job-id-1",
                "job_config_revision": "1685740490096",
                "device_config_id": "conf-id-1",
            },
        ),
        Device(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "device_id": "dev-id-2",
                "type": "device",
                "service": "compute",
                "friendly_name": "hq-orin-02",
                "hostname": "hq-orin-02",
                "status": "active",
                "online": True,
                "maintenance_mode": False,
                "job_status": "running",
                "job_last_deployed": "2023-06-02T21:22:40.084Z",
                "job_config_id": "job-id-2",
                "job_config_revision": "1685740490096",
                "device_config_id": "conf-id-1",
            },
        ),
    ]


@fixture(scope="session")
def mock_datastores():
    proximl = Mock()
    yield [
        Datastore(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "store_id": "store-id-1",
                "type": "nfs",
                "name": "On-prem NFS",
                "uri": "192.168.0.50",
                "root": "/exports",
            },
        ),
        Datastore(
            proximl,
            **{
                "provider_uuid": "prov-id-2",
                "region_uuid": "reg-id-2",
                "store_id": "store-id-2",
                "type": "smb",
                "name": "GCP Samba",
                "uri": "192.168.1.50",
                "root": "/DATA",
            },
        ),
    ]


@fixture(scope="session")
def mock_project_datastores():
    proximl = Mock()
    yield [
        ProjectDatastore(
            proximl,
            **{
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-1",
                "id": "store-id-1",
                "type": "nfs",
                "name": "On-prem NFS",
            },
        ),
        ProjectDatastore(
            proximl,
            **{
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-2",
                "id": "store-id-2",
                "type": "smb",
                "name": "GCP Samba",
            },
        ),
    ]


@fixture(scope="session")
def mock_reservations():
    proximl = Mock()
    yield [
        Reservation(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "reservation_id": "res-id-1",
                "type": "port",
                "name": "On-Prem Service A",
                "resource": "8001",
                "hostname": "service-a.local",
            },
        ),
        Reservation(
            proximl,
            **{
                "provider_uuid": "prov-id-2",
                "region_uuid": "reg-id-2",
                "reservation_id": "res-id-2",
                "type": "port",
                "name": "Cloud Service B",
                "resource": "8001",
                "hostname": "service-b.local",
            },
        ),
    ]


@fixture(scope="session")
def mock_project_reservations():
    proximl = Mock()
    yield [
        ProjectReservation(
            proximl,
            **{
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-1",
                "id": "res-id-1",
                "type": "port",
                "name": "On-Prem Service A",
                "resource": "8001",
                "hostname": "service-a.local",
            },
        ),
        ProjectReservation(
            proximl,
            **{
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-2",
                "id": "res-id-2",
                "type": "port",
                "name": "Cloud Service B",
                "resource": "8001",
                "hostname": "service-b.local",
            },
        ),
    ]


@fixture(scope="session")
def mock_device_configs():
    proximl = Mock()
    yield [
        DeviceConfig(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "config_id": "conf-id-1",
                "name": "IoT 1",
            },
        ),
        DeviceConfig(
            proximl,
            **{
                "provider_uuid": "prov-id-1",
                "region_uuid": "reg-id-1",
                "config_id": "conf-id-2",
                "name": "IoT 2",
            },
        ),
    ]


@fixture(scope="function")
def mock_proximl(
    mock_my_datasets,
    mock_public_datasets,
    mock_models,
    mock_gpu_types,
    mock_environments,
    mock_jobs,
    mock_projects,
    mock_providers,
    mock_regions,
    mock_nodes,
    mock_devices,
    mock_datastores,
    mock_reservations,
    mock_device_configs,
):
    proximl = create_autospec(ProxiML)
    proximl.active_project = "proj-id-1"
    proximl.project = "proj-id-1"
    proximl.datasets = create_autospec(Datasets)
    proximl.checkpoints = create_autospec(Checkpoints)
    proximl.models = create_autospec(Models)
    proximl.gpu_types = create_autospec(GpuTypes)
    proximl.environments = create_autospec(Environments)
    proximl.jobs = create_autospec(Jobs)
    proximl.connections = create_autospec(Connections)
    proximl.projects = create_autospec(Projects)
    proximl.datasets.list = AsyncMock(return_value=mock_my_datasets)
    proximl.datasets.list_public = AsyncMock(return_value=mock_public_datasets)
    proximl.checkpoints.list = AsyncMock(return_value=mock_my_checkpoints)
    proximl.checkpoints.list_public = AsyncMock(
        return_value=mock_public_checkpoints
    )
    proximl.models.list = AsyncMock(return_value=mock_models)
    proximl.gpu_types.list = AsyncMock(return_value=mock_gpu_types)
    proximl.environments.list = AsyncMock(return_value=mock_environments)
    proximl.jobs.list = AsyncMock(return_value=mock_jobs)
    proximl.projects.list = AsyncMock(return_value=mock_projects)

    proximl.cloudbender = create_autospec(Cloudbender)

    proximl.cloudbender.providers = create_autospec(Providers)
    proximl.cloudbender.providers.list = AsyncMock(return_value=mock_providers)
    proximl.cloudbender.regions = create_autospec(Regions)
    proximl.cloudbender.regions.list = AsyncMock(return_value=mock_regions)
    proximl.cloudbender.nodes = create_autospec(Nodes)
    proximl.cloudbender.nodes.list = AsyncMock(return_value=mock_nodes)
    proximl.cloudbender.devices = create_autospec(Nodes)
    proximl.cloudbender.devices.list = AsyncMock(return_value=mock_devices)
    proximl.cloudbender.datastores = create_autospec(Datastores)
    proximl.cloudbender.datastores.list = AsyncMock(
        return_value=mock_datastores
    )
    proximl.cloudbender.reservations = create_autospec(Reservations)
    proximl.cloudbender.reservations.list = AsyncMock(
        return_value=mock_reservations
    )
    proximl.cloudbender.device_configs = create_autospec(DeviceConfigs)
    proximl.cloudbender.device_configs.list = AsyncMock(
        return_value=mock_device_configs
    )
    yield proximl



import click
from proximl.cli import cli, pass_config, search_by_id_name
from proximl.cli.cloudbender import cloudbender


@cloudbender.group()
@pass_config
def region(config):
    """proxiML CloudBender region commands."""
    pass


@region.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID to list regions for.",
)
@pass_config
def list(config, provider):
    """List regions."""
    data = [
        ["ID", "NAME", "TYPE", "STATUS"],
        [
            "-" * 80,
            "-" * 80,
            "-" * 80,
            "-" * 80,
        ],
    ]

    regions = config.proximl.run(
        config.proximl.client.cloudbender.regions.list(provider_uuid=provider)
    )

    for region in regions:
        data.append([region.id, region.name, region.type, region.status])

    for row in data:
        click.echo(
            "{: >38.36} {: >30.28} {: >15.13} {: >15.13}" "".format(*row),
            file=config.stdout,
        )


@region.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID to create the region in.",
)
@click.option(
    "--public/--no-public",
    default=True,
    show_default=True,
    help="Nodes can run other network jobs.",
)
@click.option(
    "--storage-mode",
    "-s",
    type=click.Choice(
        [
            "local",
            "central",
        ],
        case_sensitive=False,
    ),
    help="(Physical Regions Only) The type of storage configuration the region will use.",
)
@click.option(
    "--instance-class",
    "-s",
    type=click.Choice(
        ["micro", "small", "medium", "large", "xlarge"],
        case_sensitive=False,
    ),
    help="(Cloud Regions Only) The size of the storage node.",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def create(config, provider, public, storage_mode, instance_class, name):
    """
    Creates a region.
    """
    storage = dict(storage_mode=storage_mode, instance_class=instance_class)
    return config.proximl.run(
        config.proximl.client.cloudbender.regions.create(
            provider_uuid=provider, name=name, public=public, storage=storage
        )
    )


@region.command()
@click.option(
    "--provider",
    "-p",
    type=click.STRING,
    required=True,
    help="The provider ID to delete region from.",
)
@click.argument("region", type=click.STRING)
@pass_config
def remove(config, provider, region):
    """
    Remove a region.

    REGION may be specified by name or ID, but ID is preferred.
    """
    regions = config.proximl.run(
        config.proximl.client.cloudbender.regions.list(provider_uuid=provider)
    )

    found = search_by_id_name(region, regions)
    if None is found:
        raise click.UsageError("Cannot find specified region.")

    return config.proximl.run(found.remove())



import re
import json
import click
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises

pytestmark = [mark.cli, mark.unit, mark.checkpoints]

from proximl.cli import checkpoint as specimen
from proximl.checkpoints import Checkpoint


def test_list(runner, mock_my_checkpoints):
    with patch("proximl.cli.ProxiML", new=AsyncMock) as mock_proximl:
        mock_proximl.checkpoints = AsyncMock()
        mock_proximl.checkpoints.list = AsyncMock(
            return_value=mock_my_checkpoints
        )
        result = runner.invoke(specimen, ["list"])
        print(result)
        assert result.exit_code == 0
        mock_proximl.checkpoints.list.assert_called_once()



import click
import logging
from proximl.cli import cli, pass_config, search_by_id_name


def pretty_size(num):
    if not num:
        num = 0.0
    s = ("  B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
    n = 0
    while num > 1023:
        num = num / 1024
        n += 1
    return f"{num:.2f} {s[n]}"


@cli.group()
@pass_config
def model(config):
    """proxiML model commands."""
    pass


@model.command()
@click.argument("model", type=click.STRING)
@pass_config
def attach(config, model):
    """
    Attach to model and show creation logs.

    MODEL may be specified by name or ID, but ID is preferred.
    """
    models = config.proximl.run(config.proximl.client.models.list())

    found = search_by_id_name(model, models)
    if None is found:
        raise click.UsageError("Cannot find specified model.")

    try:
        config.proximl.run(found.attach())
        return config.proximl.run(found.disconnect())
    except:
        try:
            config.proximl.run(found.disconnect())
        except:
            pass
        raise


@model.command()
@click.option(
    "--attach/--no-attach",
    default=True,
    show_default=True,
    help="Auto attach to model and show creation logs.",
)
@click.argument("model", type=click.STRING)
@pass_config
def connect(config, model, attach):
    """
    Connect local source to model and begin upload.

    MODEL may be specified by name or ID, but ID is preferred.
    """
    models = config.proximl.run(config.proximl.client.models.list())

    found = search_by_id_name(model, models)
    logging.debug(found)
    if None is found:
        raise click.UsageError("Cannot find specified model.")

    try:
        if attach:
            config.proximl.run(found.connect(), found.attach())
            return config.proximl.run(found.disconnect())
        else:
            return config.proximl.run(found.connect())
    except:
        try:
            config.proximl.run(found.disconnect())
        except:
            pass
        raise


@model.command()
@click.option(
    "--attach/--no-attach",
    default=True,
    show_default=True,
    help="Auto attach to model and show creation logs.",
)
@click.option(
    "--connect/--no-connect",
    default=True,
    show_default=True,
    help="Auto connect source and start model creation.",
)
@click.option(
    "--source",
    "-s",
    type=click.Choice(["local"], case_sensitive=False),
    default="local",
    show_default=True,
    help="Dataset source type.",
)
@click.argument("name", type=click.STRING)
@click.argument(
    "path", type=click.Path(exists=True, file_okay=False, resolve_path=True)
)
@pass_config
def create(config, attach, connect, source, name, path):
    """
    Create a model.

    A model is created with the specified NAME using a local source at the PATH
    specified. PATH should be a local directory containing the source data for
    a local source or a URI for all other source types.
    """

    if source == "local":
        model = config.proximl.run(
            config.proximl.client.models.create(
                name=name, source_type="local", source_uri=path
            )
        )

        try:
            if connect and attach:
                config.proximl.run(model.attach(), model.connect())
                return config.proximl.run(model.disconnect())
            elif connect:
                return config.proximl.run(model.connect())
            else:
                raise click.UsageError(
                    "Abort!\n"
                    "No logs to show for local sourced model without connect."
                )
        except:
            try:
                config.proximl.run(model.disconnect())
            except:
                pass
            raise


@model.command()
@click.argument("model", type=click.STRING)
@pass_config
def disconnect(config, model):
    """
    Disconnect and clean-up model upload.

    MODEL may be specified by name or ID, but ID is preferred.
    """
    models = config.proximl.run(config.proximl.client.models.list())

    found = search_by_id_name(model, models)
    if None is found:
        raise click.UsageError("Cannot find specified model.")

    return config.proximl.run(found.disconnect())


@model.command()
@pass_config
def list(config):
    """List models."""
    data = [
        ["ID", "STATUS", "NAME", "SIZE"],
        ["-" * 80, "-" * 80, "-" * 80, "-" * 80],
    ]

    models = config.proximl.run(config.proximl.client.models.list())

    for model in models:
        data.append(
            [
                model.id,
                model.status,
                model.name,
                pretty_size(model.size),
            ]
        )
    for row in data:
        click.echo(
            "{: >38.36} {: >13.11} {: >40.38} {: >14.12}" "".format(*row),
            file=config.stdout,
        )


@model.command()
@click.option(
    "--force/--no-force",
    default=False,
    show_default=True,
    help="Force removal.",
)
@click.argument("model", type=click.STRING)
@pass_config
def remove(config, model, force):
    """
    Remove a model.

    MODEL may be specified by name or ID, but ID is preferred.
    """
    models = config.proximl.run(config.proximl.client.models.list())

    found = search_by_id_name(model, models)
    if None is found:
        if force:
            config.proximl.run(found.client.models.remove(model))
        else:
            raise click.UsageError("Cannot find specified model.")

    return config.proximl.run(found.remove(force=force))


@model.command()
@click.argument("model", type=click.STRING)
@click.argument("name", type=click.STRING)
@pass_config
def rename(config, model, name):
    """
    Renames a model.

    MODEL may be specified by name or ID, but ID is preferred.
    """
    try:
        model = config.proximl.run(config.proximl.client.models.get(model))
        if model is None:
            raise click.UsageError("Cannot find specified model.")
    except:
        raise click.UsageError("Cannot find specified model.")

    return config.proximl.run(model.rename(name=name))



import re
import json
import logging
from unittest.mock import AsyncMock, patch
from pytest import mark, fixture, raises
from aiohttp import WSMessage, WSMsgType

import proximl.projects as specimen
from proximl.exceptions import (
    ApiError,
    SpecificationError,
    ProxiMLException,
)

pytestmark = [mark.sdk, mark.unit, mark.projects]


@fixture
def projects(mock_proximl):
    yield specimen.Projects(mock_proximl)


@fixture
def project(mock_proximl):
    yield specimen.Project(
        mock_proximl,
        id="1",
        name="My Mock Project",
        owner=True,
        owner_name="Me",
        created_name="Me",
        job_all=True,
        dataset_all=True,
        model_all=True,
        createdAt="2020-12-31T23:59:59.000Z",
    )


@fixture
def project_datastore(mock_proximl):
    yield specimen.ProjectDatastore(
        mock_proximl,
        id="ds-id-1",
        name="datastore 1",
        project_uuid="proj-id-1",
        type="nfs",
        region_uuid="reg-id-1",
    )


@fixture
def project_reservation(mock_proximl):
    yield specimen.ProjectReservation(
        mock_proximl,
        id="res-id-1",
        name="reservation 1",
        project_uuid="proj-id-1",
        region_uuid="reg-id-1",
        type="port",
        resource="8001",
        hostname="service.local",
    )


class ProjectsTests:
    @mark.asyncio
    async def test_get_project(
        self,
        projects,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await projects.get("1234")
        mock_proximl._query.assert_called_once_with(
            "/project/1234", "GET", dict()
        )

    @mark.asyncio
    async def test_list_projects(
        self,
        projects,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await projects.list()
        mock_proximl._query.assert_called_once_with("/project", "GET", dict())

    @mark.asyncio
    async def test_remove_project(
        self,
        projects,
        mock_proximl,
    ):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await projects.remove("4567")
        mock_proximl._query.assert_called_once_with(
            "/project/4567", "DELETE", dict()
        )

    @mark.asyncio
    async def test_create_project_simple(self, projects, mock_proximl):
        requested_config = dict(
            name="new project",
        )
        expected_payload = dict(name="new project", copy_keys=False)
        api_response = {
            "id": "project-id-1",
            "name": "new project",
            "owner": True,
            "owner_name": "Me",
            "created_name": "Me",
            "job_all": True,
            "dataset_all": True,
            "model_all": True,
            "createdAt": "2020-12-31T23:59:59.000Z",
        }

        mock_proximl._query = AsyncMock(return_value=api_response)
        response = await projects.create(**requested_config)
        mock_proximl._query.assert_called_once_with(
            "/project", "POST", None, expected_payload
        )
        assert response.id == "project-id-1"


class ProjectDatastoreTests:
    def test_project_datastore_properties(self, project_datastore):
        assert isinstance(project_datastore.id, str)
        assert isinstance(project_datastore.name, str)
        assert isinstance(project_datastore.project_uuid, str)
        assert isinstance(project_datastore.type, str)
        assert isinstance(project_datastore.region_uuid, str)

    def test_project_datastore_str(self, project_datastore):
        string = str(project_datastore)
        regex = r"^{.*\"id\": \"" + project_datastore.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_datastore_repr(self, project_datastore):
        string = repr(project_datastore)
        regex = (
            r"^ProjectDatastore\( proximl , \*\*{.*'id': '"
            + project_datastore.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_datastore_bool(self, project_datastore, mock_proximl):
        empty_project_datastore = specimen.ProjectDatastore(mock_proximl)
        assert bool(project_datastore)
        assert not bool(empty_project_datastore)


class ProjectReservationTests:
    def test_project_reservation_properties(self, project_reservation):
        assert isinstance(project_reservation.id, str)
        assert isinstance(project_reservation.name, str)
        assert isinstance(project_reservation.project_uuid, str)
        assert isinstance(project_reservation.type, str)
        assert isinstance(project_reservation.hostname, str)
        assert isinstance(project_reservation.resource, str)
        assert isinstance(project_reservation.region_uuid, str)

    def test_project_reservation_str(self, project_reservation):
        string = str(project_reservation)
        regex = r"^{.*\"id\": \"" + project_reservation.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_reservation_repr(self, project_reservation):
        string = repr(project_reservation)
        regex = (
            r"^ProjectReservation\( proximl , \*\*{.*'id': '"
            + project_reservation.id
            + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_reservation_bool(self, project_reservation, mock_proximl):
        empty_project_reservation = specimen.ProjectReservation(mock_proximl)
        assert bool(project_reservation)
        assert not bool(empty_project_reservation)


class ProjectTests:
    def test_project_properties(self, project):
        assert isinstance(project.id, str)
        assert isinstance(project.name, str)
        assert isinstance(project.owner_name, str)
        assert isinstance(project.is_owner, bool)

    def test_project_str(self, project):
        string = str(project)
        regex = r"^{.*\"id\": \"" + project.id + r"\".*}$"
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_repr(self, project):
        string = repr(project)
        regex = (
            r"^Project\( proximl , \*\*{.*'id': '" + project.id + r"'.*}\)$"
        )
        assert isinstance(string, str)
        assert re.match(regex, string)

    def test_project_bool(self, project, mock_proximl):
        empty_project = specimen.Project(mock_proximl)
        assert bool(project)
        assert not bool(empty_project)

    @mark.asyncio
    async def test_project_remove(self, project, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project.remove()
        mock_proximl._query.assert_called_once_with("/project/1", "DELETE")

    @mark.asyncio
    async def test_project_refresh_datastores(self, project, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project.refresh_datastores()
        mock_proximl._query.assert_called_once_with(
            "/project/1/datastores", "PATCH"
        )

    @mark.asyncio
    async def test_project_refresh_reservations(self, project, mock_proximl):
        api_response = dict()
        mock_proximl._query = AsyncMock(return_value=api_response)
        await project.refresh_reservations()
        mock_proximl._query.assert_called_once_with(
            "/project/1/reservations", "PATCH"
        )

    @mark.asyncio
    async def test_project_list_datastores(self, project, mock_proximl):
        api_response = [
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-1",
                "id": "store-id-1",
                "type": "nfs",
                "name": "On-prem NFS",
            },
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-2",
                "id": "store-id-2",
                "type": "smb",
                "name": "GCP Samba",
            },
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project.list_datastores()
        mock_proximl._query.assert_called_once_with(
            "/project/1/datastores", "GET"
        )
        assert len(resp) == 2

    @mark.asyncio
    async def test_project_list_reservations(self, project, mock_proximl):
        api_response = [
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-1",
                "id": "res-id-1",
                "type": "port",
                "name": "On-Prem Service A",
                "resource": "8001",
                "hostname": "service-a.local",
            },
            {
                "project_uuid": "proj-id-1",
                "region_uuid": "reg-id-2",
                "id": "res-id-2",
                "type": "port",
                "name": "Cloud Service B",
                "resource": "8001",
                "hostname": "service-b.local",
            },
        ]
        mock_proximl._query = AsyncMock(return_value=api_response)
        resp = await project.list_reservations()
        mock_proximl._query.assert_called_once_with(
            "/project/1/reservations", "GET"
        )
        assert len(resp) == 2



import click
import json
from webbrowser import open as browse
from proximl.cli import pass_config
from proximl.cli.job import job


def validate_cpu_count(ctx, param, value):
    if value is None:
        return
    try:
        count = int(value)
    except:
        raise click.BadParameter("must be integer")
    if count % 4 != 0:
        raise click.BadParameter("must be multiple of 4")

    return count


@job.group()
@pass_config
def create(config):
    """proxiML job create."""
    pass


@create.command()
@click.option(
    "--attach/--no-attach",
    default=True,
    show_default=True,
    help="Auto attach to job.",
)
@click.option(
    "--connect/--no-connect",
    default=True,
    show_default=True,
    help="Auto connect to job.",
)
@click.option(
    "--disk-size",
    "-ds",
    type=click.INT,
    default=10,
    show_default=True,
    help="Disk size (GiB).",
)
@click.option(
    "--gpu-count",
    "-gc",
    type=click.INT,
    default=1,
    show_default=True,
    help="GPU Count (per Worker).",
)
@click.option(
    "--gpu-type",
    "-gt",
    type=click.Choice(
        [
            "gtx1060",
            "gtx1080ti",
            "rtx2060s",
            "rtx2070s",
            "rtx2080ti",
            "rtx3090",
            "p100",
            "t4",
            "v100",
            "a100",
            "a100xl",
            "a6000",
            "a4000",
            "cpu",
        ],
        case_sensitive=False,
    ),
    default=["rtx3090"],
    show_default=True,
    multiple=True,
    help="GPU type.",
)
@click.option(
    "--cpu-count",
    "-cc",
    type=click.UNPROCESSED,
    callback=validate_cpu_count,
    help="CPU Count (per Worker).",
)
@click.option(
    "--max-price",
    "-mp",
    type=click.FLOAT,
    default=10.0,
    show_default=True,
    help="Max Price (per GPU).",
)
@click.option(
    "--data-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to copy as the input data",
)
@click.option(
    "--dataset",
    type=click.STRING,
    help="ID or Name of a dataset to add to the job",
    multiple=True,
)
@click.option(
    "--public-dataset",
    type=click.STRING,
    help="ID or Name of a public dataset to add to the job",
    multiple=True,
)
@click.option(
    "--checkpoint",
    type=click.STRING,
    help="ID or Name of a checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--public-checkpoint",
    type=click.STRING,
    help="ID or Name of a public checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--environment",
    type=click.Choice(
        [
            "DEEPLEARNING_PY39",
            "DEEPLEARNING_PY310",
            "PYTORCH_PY39_20",
            "PYTORCH_PY39_113",
            "PYTORCH_PY39_112",
            "TENSORFLOW_PY39_212",
            "TENSORFLOW_PY39_211",
            "TENSORFLOW_PY39_210",
            "MXNET_PY39_19",
        ],
        case_sensitive=False,
    ),
    default="DEEPLEARNING_PY310",
    show_default=True,
    help="Job environment to use",
)
@click.option(
    "--custom-image",
    type=click.STRING,
    help="Docker Image to use for the job.  Implies 'CUSTOM' environment type.",
)
@click.option(
    "--env",
    type=click.STRING,
    help="Environment variables to set in the job environment in 'KEY=VALUE' format",
    multiple=True,
)
@click.option(
    "--apt-packages",
    type=click.STRING,
    help="Apt packages to install as a comma separated list 'p1,p2=v2,p3'",
)
@click.option(
    "--pip-packages",
    type=click.STRING,
    help="Pip packages to install as a comma separated list 'p1,p2==v2,p3'",
)
@click.option(
    "--conda-packages",
    type=click.STRING,
    help="Conda packages to install as a comma separated list 'p1,\"p2=v2\",p3'",
)
@click.option(
    "--key",
    type=click.Choice(
        ["aws", "gcp", "kaggle", "azure", "wasabi"],
        case_sensitive=False,
    ),
    help="Third Party Keys to add to the job environment",
    multiple=True,
)
@click.option(
    "--git-uri",
    type=click.STRING,
    help="Git repository to use as the model data",
)
@click.option(
    "--model-id",
    type=click.STRING,
    help="proxiML Model ID to use as the model data",
)
@click.option(
    "--model-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to copy as the model data",
)
@click.option(
    "--timeout",
    type=click.INT,
    default=300,
    show_default=True,
    help="Seconds to wait for job to start",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def notebook(
    config,
    attach,
    connect,
    disk_size,
    gpu_count,
    gpu_type,
    cpu_count,
    max_price,
    data_dir,
    dataset,
    public_dataset,
    checkpoint,
    public_checkpoint,
    environment,
    custom_image,
    env,
    key,
    apt_packages,
    pip_packages,
    conda_packages,
    model_dir,
    git_uri,
    model_id,
    timeout,
    name,
):
    """
    Create a notebook.
    """

    datasets = [
        dict(
            id=item,
        )
        for item in dataset
    ] + [dict(id=item, public=True) for item in public_dataset]

    checkpoints = [
        dict(
            id=item,
        )
        for item in checkpoint
    ] + [dict(id=item, public=True) for item in public_checkpoint]

    options = dict(
        max_price=max_price,
        data=dict(datasets=datasets),
        model=dict(checkpoints=checkpoints),
        environment=dict(
            worker_key_types=[k for k in key],
        ),
    )

    if not (len(gpu_type) == 1 and gpu_type[0] == "cpu"):
        options["gpu_count"] = gpu_count

    if custom_image:
        options["environment"]["type"] = "CUSTOM"
        options["environment"]["custom_image"] = custom_image
    else:
        options["environment"]["type"] = environment

    try:
        envs = [
            {"key": e.split("=")[0], "value": e.split("=")[1]} for e in env
        ]
        options["environment"]["env"] = envs
    except IndexError:
        raise click.UsageError(
            "Invalid environment variable format.  Must be in 'KEY=VALUE' format."
        )

    if apt_packages or pip_packages or conda_packages:
        options["environment"]["packages"] = dict()
        if apt_packages:
            options["environment"]["packages"]["apt"] = apt_packages.split(",")
        if pip_packages:
            options["environment"]["packages"]["pip"] = pip_packages.split(",")
        if conda_packages:
            options["environment"]["packages"]["conda"] = conda_packages.split(
                ","
            )

    if data_dir:
        click.echo("Creating Dataset..", file=config.stdout)
        new_dataset = config.proximl.run(
            config.proximl.client.datasets.create(
                f"Job - {name}", "local", data_dir
            )
        )
        if attach:
            config.proximl.run(new_dataset.attach(), new_dataset.connect())
            config.proximl.run(new_dataset.disconnect())
        else:
            config.proximl.run(new_dataset.connect())
            config.proximl.run(new_dataset.wait_for("ready"))
            config.proximl.run(new_dataset.disconnect())
        options["data"]["datasets"].append(
            dict(id=new_dataset.id, type="existing")
        )

    if git_uri:
        options["model"]["source_type"] = "git"
        options["model"]["source_uri"] = git_uri
    if model_id:
        options["model"]["source_type"] = "proximl"
        options["model"]["source_uri"] = model_id
    if model_dir:
        options["model"]["source_type"] = "local"
        options["model"]["source_uri"] = model_dir

    job = config.proximl.run(
        config.proximl.client.jobs.create(
            name=name,
            type="notebook",
            gpu_types=gpu_type,
            cpu_count=cpu_count,
            disk_size=disk_size,
            **options,
        )
    )
    click.echo("Created Job.", file=config.stdout)
    if model_dir:
        config.proximl.run(job.wait_for("waiting for data/model download"))
        if attach or connect:
            click.echo("Waiting for job to start...", file=config.stdout)
            config.proximl.run(job.connect(), job.attach())
            config.proximl.run(job.disconnect())
            click.echo("Launching...", file=config.stdout)
            browse(job.notebook_url)
        else:
            config.proximl.run(job.connect())
            config.proximl.run(job.wait_for("running"))
            config.proximl.run(job.disconnect())
    elif attach or connect:
        click.echo("Waiting for job to start...", file=config.stdout)
        config.proximl.run(job.wait_for("running", timeout))
        click.echo("Launching...", file=config.stdout)
        config.proximl.run(job.open())


@create.command()
@click.option(
    "--attach/--no-attach",
    default=True,
    show_default=True,
    help="Auto attach to job.",
)
@click.option(
    "--connect/--no-connect",
    default=True,
    show_default=True,
    help="Auto connect to job.",
)
@click.option(
    "--disk-size",
    "-ds",
    type=click.INT,
    default=10,
    show_default=True,
    help="Disk size (GiB).",
)
@click.option(
    "--gpu-count",
    "-gc",
    type=click.INT,
    default=1,
    show_default=True,
    help="GPU Count (per Worker.)",
)
@click.option(
    "--gpu-type",
    "-gt",
    type=click.Choice(
        [
            "gtx1060",
            "gtx1080ti",
            "rtx2060s",
            "rtx2070s",
            "rtx2080ti",
            "rtx3090",
            "p100",
            "t4",
            "v100",
            "a100",
            "a100xl",
            "a6000",
            "a4000",
            "cpu",
        ],
        case_sensitive=False,
    ),
    default="rtx3090",
    show_default=True,
    help="GPU type.",
)
@click.option(
    "--cpu-count",
    "-cc",
    type=click.UNPROCESSED,
    callback=validate_cpu_count,
    help="CPU Count (per Worker).",
)
@click.option(
    "--max-price",
    "-mp",
    type=click.FLOAT,
    default=10.0,
    show_default=True,
    help="Max Price (per GPU).",
)
@click.option(
    "--data-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to copy as the input data",
)
@click.option(
    "--dataset",
    type=click.STRING,
    help="ID or Name of a dataset to add to the job",
    multiple=True,
)
@click.option(
    "--public-dataset",
    type=click.STRING,
    help="ID or Name of a public dataset to add to the job",
    multiple=True,
)
@click.option(
    "--checkpoint",
    type=click.STRING,
    help="ID or Name of a checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--public-checkpoint",
    type=click.STRING,
    help="ID or Name of a public checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--output-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to save the output results to",
)
@click.option(
    "--output-type",
    type=click.Choice(
        ["aws", "gcp", "local", "proximl", "azure", "wasabi"],
        case_sensitive=False,
    ),
    help="Service provider to store the output data",
)
@click.option(
    "--output-uri",
    type=click.STRING,
    help="Location in the output-type provider to store the output data",
)
@click.option(
    "--archive/--no-archive",
    default=False,
    show_default=True,
    help="Zip the output contents before uploading.",
)
@click.option(
    "--save-model/--no-save-model",
    default=False,
    show_default=True,
    help="Include the model directory in the job output",
)
@click.option(
    "--environment",
    type=click.Choice(
        [
            "DEEPLEARNING_PY39",
            "DEEPLEARNING_PY310",
            "PYTORCH_PY39_20",
            "PYTORCH_PY39_113",
            "PYTORCH_PY39_112",
            "TENSORFLOW_PY39_212",
            "TENSORFLOW_PY39_211",
            "TENSORFLOW_PY39_210",
            "MXNET_PY39_19",
        ],
        case_sensitive=False,
    ),
    default="DEEPLEARNING_PY310",
    show_default=True,
    help="Job environment to use",
)
@click.option(
    "--custom-image",
    type=click.STRING,
    help="Docker Image to use for the job.  Implies 'CUSTOM' environment type.",
)
@click.option(
    "--env",
    type=click.STRING,
    help="Environment variables to set in the job environment in 'KEY=VALUE' format",
    multiple=True,
)
@click.option(
    "--key",
    type=click.Choice(
        ["aws", "gcp", "kaggle", "azure", "wasabi"],
        case_sensitive=False,
    ),
    help="Third Party Keys to add to the job environment",
    multiple=True,
)
@click.option(
    "--apt-packages",
    type=click.STRING,
    help="Apt packages to install as a comma separated list 'p1,p2=v2,p3'",
)
@click.option(
    "--pip-packages",
    type=click.STRING,
    help="Pip packages to install as a comma separated list 'p1,p2==v2,p3'",
)
@click.option(
    "--conda-packages",
    type=click.STRING,
    help="Conda packages to install as a comma separated list 'p1,\"p2=v2\",p3'",
)
@click.option(
    "--git-uri",
    type=click.STRING,
    help="Git repository to use as the model data",
)
@click.option(
    "--model-id",
    type=click.STRING,
    help="proxiML Model ID to use as the model data",
)
@click.option(
    "--model-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to copy as the model data",
)
@click.argument("name", type=click.STRING, required=True)
@click.argument("commands", type=click.STRING, nargs=-1, required=True)
@pass_config
def training(
    config,
    attach,
    connect,
    disk_size,
    gpu_count,
    gpu_type,
    cpu_count,
    max_price,
    data_dir,
    dataset,
    public_dataset,
    checkpoint,
    public_checkpoint,
    output_dir,
    output_type,
    output_uri,
    archive,
    save_model,
    environment,
    custom_image,
    env,
    key,
    apt_packages,
    pip_packages,
    conda_packages,
    model_dir,
    git_uri,
    model_id,
    name,
    commands,
):
    """
    Create a training job.
    """

    datasets = [
        dict(
            id=item,
        )
        for item in dataset
    ] + [dict(id=item, public=True) for item in public_dataset]

    checkpoints = [
        dict(
            id=item,
        )
        for item in checkpoint
    ] + [dict(id=item, public=True) for item in public_checkpoint]

    options = dict(
        max_price=max_price,
        data=dict(datasets=datasets),
        model=dict(checkpoints=checkpoints),
        environment=dict(
            worker_key_types=[k for k in key],
        ),
    )

    if not (len(gpu_type) == 1 and gpu_type[0] == "cpu"):
        options["gpu_count"] = gpu_count

    if custom_image:
        options["environment"]["type"] = "CUSTOM"
        options["environment"]["custom_image"] = custom_image
    else:
        options["environment"]["type"] = environment

    if output_type:
        options["data"]["output_type"] = output_type
        options["data"]["output_uri"] = output_uri
        options["data"]["output_options"] = dict(
            archive=archive, save_model=save_model
        )

    if output_dir:
        options["data"]["output_type"] = "local"
        options["data"]["output_uri"] = output_dir
        options["data"]["output_options"] = dict(
            archive=archive, save_model=save_model
        )

    try:
        envs = [
            {"key": e.split("=")[0], "value": e.split("=")[1]} for e in env
        ]
        options["environment"]["env"] = envs
    except IndexError:
        raise click.UsageError(
            "Invalid environment variable format.  Must be in 'KEY=VALUE' format."
        )

    if apt_packages or pip_packages or conda_packages:
        options["environment"]["packages"] = dict()
        if apt_packages:
            options["environment"]["packages"]["apt"] = apt_packages.split(",")
        if pip_packages:
            options["environment"]["packages"]["pip"] = pip_packages.split(",")
        if conda_packages:
            options["environment"]["packages"]["conda"] = conda_packages.split(
                ","
            )

    if data_dir:
        click.echo("Creating Dataset..", file=config.stdout)
        new_dataset = config.proximl.run(
            config.proximl.client.datasets.create(
                f"Job - {name}", "local", data_dir
            )
        )
        if attach:
            config.proximl.run(new_dataset.attach(), new_dataset.connect())
            config.proximl.run(new_dataset.disconnect())
        else:
            config.proximl.run(new_dataset.connect())
            config.proximl.run(new_dataset.wait_for("ready"))
            config.proximl.run(new_dataset.disconnect())
        options["data"]["datasets"].append(
            dict(id=new_dataset.id, type="existing")
        )

    if git_uri:
        options["model"]["source_type"] = "git"
        options["model"]["source_uri"] = git_uri
    if model_id:
        options["model"]["source_type"] = "proximl"
        options["model"]["source_uri"] = model_id
    if model_dir:
        options["model"]["source_type"] = "local"
        options["model"]["source_uri"] = model_dir

    job = config.proximl.run(
        config.proximl.client.jobs.create(
            name=name,
            type="training",
            gpu_types=gpu_type,
            cpu_count=cpu_count,
            disk_size=disk_size,
            workers=[command for command in commands],
            **options,
        )
    )
    click.echo("Created Job.", file=config.stdout)

    if connect or attach:
        config.proximl.run(job.wait_for("waiting for data/model download"))
        if connect:
            config.proximl.run(job.connect())
        if attach:
            config.proximl.run(job.attach())


@create.command()
@click.option(
    "--attach/--no-attach",
    default=True,
    show_default=True,
    help="Auto attach to job.",
)
@click.option(
    "--connect/--no-connect",
    default=True,
    show_default=True,
    help="Auto connect to job.",
)
@click.option(
    "--disk-size",
    "-ds",
    type=click.INT,
    default=10,
    show_default=True,
    help="Disk size (GiB).",
)
@click.option(
    "--gpu-count",
    "-gc",
    type=click.INT,
    default=1,
    show_default=True,
    help="GPU Count (per Worker.)",
)
@click.option(
    "--gpu-type",
    "-gt",
    type=click.Choice(
        [
            "gtx1060",
            "gtx1080ti",
            "rtx2060s",
            "rtx2070s",
            "rtx2080ti",
            "rtx3090",
            "p100",
            "t4",
            "v100",
            "a100",
            "a100xl",
            "a6000",
            "a4000",
            "cpu",
        ],
        case_sensitive=False,
    ),
    default="rtx3090",
    show_default=True,
    help="GPU type.",
)
@click.option(
    "--cpu-count",
    "-cc",
    type=click.UNPROCESSED,
    callback=validate_cpu_count,
    help="CPU Count (per Worker).",
)
@click.option(
    "--max-price",
    "-mp",
    type=click.FLOAT,
    default=10.0,
    show_default=True,
    help="Max Price (per GPU).",
)
@click.option(
    "--checkpoint",
    type=click.STRING,
    help="ID or Name of a checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--public-checkpoint",
    type=click.STRING,
    help="ID or Name of a public checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--input-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to copy as the input data",
)
@click.option(
    "--input-type",
    type=click.Choice(
        ["aws", "gcp", "local", "kaggle", "web", "azure", "wasabi"],
        case_sensitive=False,
    ),
    help="Service provider to load the input data",
)
@click.option(
    "--input-uri",
    type=click.STRING,
    help="Location in the input-type provider of the input data",
)
@click.option(
    "--output-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to save the output results to",
)
@click.option(
    "--output-type",
    type=click.Choice(
        ["aws", "gcp", "local", "proximl", "azure", "wasabi"],
        case_sensitive=False,
    ),
    help="Service provider to store the output data",
)
@click.option(
    "--output-uri",
    type=click.STRING,
    help="Location in the output-type provider to store the output data",
)
@click.option(
    "--archive/--no-archive",
    default=False,
    show_default=True,
    help="Zip the output contents before uploading.",
)
@click.option(
    "--save-model/--no-save-model",
    default=False,
    show_default=True,
    help="Include the model directory in the job output",
)
@click.option(
    "--environment",
    type=click.Choice(
        [
            "DEEPLEARNING_PY39",
            "DEEPLEARNING_PY310",
            "PYTORCH_PY39_20",
            "PYTORCH_PY39_113",
            "PYTORCH_PY39_112",
            "TENSORFLOW_PY39_212",
            "TENSORFLOW_PY39_211",
            "TENSORFLOW_PY39_210",
            "MXNET_PY39_19",
        ],
        case_sensitive=False,
    ),
    default="DEEPLEARNING_PY310",
    show_default=True,
    help="Job environment to use",
)
@click.option(
    "--custom-image",
    type=click.STRING,
    help="Docker Image to use for the job.  Implies 'CUSTOM' environment type.",
)
@click.option(
    "--env",
    type=click.STRING,
    help="Environment variables to set in the job environment in 'KEY=VALUE' format",
    multiple=True,
)
@click.option(
    "--key",
    type=click.Choice(
        ["aws", "gcp", "kaggle", "azure", "wasabi"],
        case_sensitive=False,
    ),
    help="Third Party Keys to add to the job environment",
    multiple=True,
)
@click.option(
    "--apt-packages",
    type=click.STRING,
    help="Apt packages to install as a comma separated list 'p1,p2=v2,p3'",
)
@click.option(
    "--pip-packages",
    type=click.STRING,
    help="Pip packages to install as a comma separated list 'p1,p2==v2,p3'",
)
@click.option(
    "--conda-packages",
    type=click.STRING,
    help="Conda packages to install as a comma separated list 'p1,\"p2=v2\",p3'",
)
@click.option(
    "--git-uri",
    type=click.STRING,
    help="Git repository to use as the model data",
)
@click.option(
    "--model-id",
    type=click.STRING,
    help="proxiML Model ID to use as the model data",
)
@click.option(
    "--model-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to copy as the model data",
)
@click.argument("name", type=click.STRING, required=True)
@click.argument("command", type=click.STRING, required=True)
@pass_config
def inference(
    config,
    attach,
    connect,
    disk_size,
    gpu_count,
    gpu_type,
    cpu_count,
    max_price,
    checkpoint,
    public_checkpoint,
    input_dir,
    input_type,
    input_uri,
    output_dir,
    output_type,
    output_uri,
    archive,
    save_model,
    environment,
    custom_image,
    env,
    key,
    apt_packages,
    pip_packages,
    conda_packages,
    model_dir,
    git_uri,
    model_id,
    name,
    command,
):
    """
    Create an inference job.
    """

    checkpoints = [
        dict(
            id=item,
        )
        for item in checkpoint
    ] + [dict(id=item, public=True) for item in public_checkpoint]

    options = dict(
        max_price=max_price,
        data=dict(datasets=[]),
        model=dict(checkpoints=checkpoints),
        environment=dict(
            worker_key_types=[k for k in key],
        ),
    )

    if not (len(gpu_type) == 1 and gpu_type[0] == "cpu"):
        options["gpu_count"] = gpu_count

    if custom_image:
        options["environment"]["type"] = "CUSTOM"
        options["environment"]["custom_image"] = custom_image
    else:
        options["environment"]["type"] = environment

    if input_type:
        options["data"]["input_type"] = input_type
        options["data"]["input_uri"] = input_uri

    if input_dir:
        options["data"]["input_type"] = "local"
        options["data"]["input_uri"] = input_dir

    if output_type:
        options["data"]["output_type"] = output_type
        options["data"]["output_uri"] = output_uri
        options["data"]["output_options"] = dict(
            archive=archive, save_model=save_model
        )

    if output_dir:
        options["data"]["output_type"] = "local"
        options["data"]["output_uri"] = output_dir
        options["data"]["output_options"] = dict(
            archive=archive, save_model=save_model
        )

    try:
        envs = [
            {"key": e.split("=")[0], "value": e.split("=")[1]} for e in env
        ]
        options["environment"]["env"] = envs
    except IndexError:
        raise click.UsageError(
            "Invalid environment variable format.  Must be in 'KEY=VALUE' format."
        )

    if apt_packages or pip_packages or conda_packages:
        options["environment"]["packages"] = dict()
        if apt_packages:
            options["environment"]["packages"]["apt"] = apt_packages.split(",")
        if pip_packages:
            options["environment"]["packages"]["pip"] = pip_packages.split(",")
        if conda_packages:
            options["environment"]["packages"]["conda"] = conda_packages.split(
                ","
            )

    if git_uri:
        options["model"]["source_type"] = "git"
        options["model"]["source_uri"] = git_uri
    if model_id:
        options["model"]["source_type"] = "proximl"
        options["model"]["source_uri"] = model_id
    if model_dir:
        options["model"]["source_type"] = "local"
        options["model"]["source_uri"] = model_dir

    job = config.proximl.run(
        config.proximl.client.jobs.create(
            name=name,
            type="inference",
            gpu_types=gpu_type,
            cpu_count=cpu_count,
            disk_size=disk_size,
            workers=[command],
            **options,
        )
    )
    click.echo("Created Job.", file=config.stdout)

    if connect or attach:
        config.proximl.run(job.wait_for("waiting for data/model download"))
        if connect:
            config.proximl.run(job.connect())
        if attach:
            config.proximl.run(job.attach())


@create.command()
@click.option(
    "--attach/--no-attach",
    default=True,
    show_default=True,
    help="Auto attach to job.",
)
@click.option(
    "--connect/--no-connect",
    default=True,
    show_default=True,
    help="Auto connect to job.",
)
@click.argument("file", type=click.File("rb"))
@pass_config
def from_json(config, attach, connect, file):
    """
    Create an job from json file representation.
    """
    payload = json.loads(file.read())

    job = config.proximl.run(config.proximl.client.jobs.create_json(payload))
    click.echo("Created.", file=config.stdout)
    if attach or connect:
        if job.type == "notebook":
            click.echo("Waiting for job to start...", file=config.stdout)
            config.proximl.run(job.wait_for("running"))
            click.echo("Launching...", file=config.stdout)
            browse(job.notebook_url)
        else:
            if connect:
                config.proximl.run(job.connect())
            if attach:
                config.proximl.run(job.attach())


@create.command()
@click.option(
    "--attach/--no-attach",
    default=True,
    show_default=True,
    help="Auto attach to job.",
)
@click.option(
    "--connect/--no-connect",
    default=True,
    show_default=True,
    help="Auto connect to job.",
)
@click.option(
    "--disk-size",
    "-ds",
    type=click.INT,
    default=10,
    show_default=True,
    help="Disk size (GiB).",
)
@click.option(
    "--gpu-count",
    "-gc",
    type=click.INT,
    default=1,
    show_default=True,
    help="GPU Count (per Worker.)",
)
@click.option(
    "--gpu-type",
    "-gt",
    type=click.Choice(
        [
            "gtx1060",
            "gtx1080ti",
            "rtx2060s",
            "rtx2070s",
            "rtx2080ti",
            "rtx3090",
            "p100",
            "t4",
            "v100",
            "a100",
            "a100xl",
            "a6000",
            "a4000",
            "cpu",
        ],
        case_sensitive=False,
    ),
    default="rtx3090",
    show_default=True,
    help="GPU type.",
)
@click.option(
    "--cpu-count",
    "-cc",
    type=click.UNPROCESSED,
    callback=validate_cpu_count,
    help="CPU Count (per Worker).",
)
@click.option(
    "--max-price",
    "-mp",
    type=click.FLOAT,
    default=10.0,
    show_default=True,
    help="Max Price (per GPU).",
)
@click.option(
    "--checkpoint",
    type=click.STRING,
    help="ID or Name of a checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--public-checkpoint",
    type=click.STRING,
    help="ID or Name of a public checkpoint to add to the job",
    multiple=True,
)
@click.option(
    "--environment",
    type=click.Choice(
        [
            "DEEPLEARNING_PY39",
            "DEEPLEARNING_PY310",
            "PYTORCH_PY39_20",
            "PYTORCH_PY39_113",
            "PYTORCH_PY39_112",
            "TENSORFLOW_PY39_212",
            "TENSORFLOW_PY39_211",
            "TENSORFLOW_PY39_210",
            "MXNET_PY39_19",
        ],
        case_sensitive=False,
    ),
    default="DEEPLEARNING_PY310",
    show_default=True,
    help="Job environment to use",
)
@click.option(
    "--custom-image",
    type=click.STRING,
    help="Docker Image to use for the job.  Implies 'CUSTOM' environment type.",
)
@click.option(
    "--env",
    type=click.STRING,
    help="Environment variables to set in the job environment in 'KEY=VALUE' format",
    multiple=True,
)
@click.option(
    "--key",
    type=click.Choice(
        ["aws", "gcp", "kaggle", "azure", "wasabi"],
        case_sensitive=False,
    ),
    help="Third Party Keys to add to the job environment",
    multiple=True,
)
@click.option(
    "--apt-packages",
    type=click.STRING,
    help="Apt packages to install as a comma separated list 'p1,p2=v2,p3'",
)
@click.option(
    "--pip-packages",
    type=click.STRING,
    help="Pip packages to install as a comma separated list 'p1,p2==v2,p3'",
)
@click.option(
    "--conda-packages",
    type=click.STRING,
    help="Conda packages to install as a comma separated list 'p1,\"p2=v2\",p3'",
)
@click.option(
    "--git-uri",
    type=click.STRING,
    help="Git repository to use as the model data",
)
@click.option(
    "--model-id",
    type=click.STRING,
    help="proxiML Model ID to use as the model data",
)
@click.option(
    "--model-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Local file path to copy as the model data",
)
@click.option(
    "--route",
    type=click.STRING,
    help="Routes to configure in endpoint (in JSON).",
    multiple=True,
)
@click.option(
    "--timeout",
    type=click.INT,
    default=300,
    show_default=True,
    help="Seconds to wait for job to start",
)
@click.argument("name", type=click.STRING, required=True)
@pass_config
def endpoint(
    config,
    attach,
    connect,
    disk_size,
    gpu_count,
    gpu_type,
    cpu_count,
    max_price,
    checkpoint,
    public_checkpoint,
    environment,
    custom_image,
    env,
    key,
    apt_packages,
    pip_packages,
    conda_packages,
    model_dir,
    git_uri,
    model_id,
    route,
    timeout,
    name,
):
    """
    Create an endpoint.
    """
    checkpoints = [
        dict(
            id=item,
        )
        for item in checkpoint
    ] + [dict(id=item, public=True) for item in public_checkpoint]

    routes = [json.loads(item) for item in route]

    options = dict(
        max_price=max_price,
        model=dict(checkpoints=checkpoints),
        environment=dict(
            worker_key_types=[k for k in key],
        ),
    )

    if not (len(gpu_type) == 1 and gpu_type[0] == "cpu"):
        options["gpu_count"] = gpu_count

    if custom_image:
        options["environment"]["type"] = "CUSTOM"
        options["environment"]["custom_image"] = custom_image
    else:
        options["environment"]["type"] = environment

    try:
        envs = [
            {"key": e.split("=")[0], "value": e.split("=")[1]} for e in env
        ]
        options["environment"]["env"] = envs
    except IndexError:
        raise click.UsageError(
            "Invalid environment variable format.  Must be in 'KEY=VALUE' format."
        )

    if apt_packages or pip_packages or conda_packages:
        options["environment"]["packages"] = dict()
        if apt_packages:
            options["environment"]["packages"]["apt"] = apt_packages.split(",")
        if pip_packages:
            options["environment"]["packages"]["pip"] = pip_packages.split(",")
        if conda_packages:
            options["environment"]["packages"]["conda"] = conda_packages.split(
                ","
            )

    if git_uri:
        options["model"]["source_type"] = "git"
        options["model"]["source_uri"] = git_uri
    if model_id:
        options["model"]["source_type"] = "proximl"
        options["model"]["source_uri"] = model_id
    if model_dir:
        options["model"]["source_type"] = "local"
        options["model"]["source_uri"] = model_dir

    job = config.proximl.run(
        config.proximl.client.jobs.create(
            name=name,
            type="endpoint",
            gpu_types=gpu_type,
            cpu_count=cpu_count,
            disk_size=disk_size,
            endpoint=dict(routes=routes),
            **options,
        )
    )
    click.echo("Created Job.", file=config.stdout)

    if model_dir:
        click.echo("Uploading model...", file=config.stdout)
        config.proximl.run(job.wait_for("waiting for data/model download"))
        if attach or connect:
            config.proximl.run(job.connect(), job.attach())
        else:
            config.proximl.run(job.connect())
        click.echo("Waiting for job to start...", file=config.stdout)
        config.proximl.run(job.wait_for("running", timeout))
        config.proximl.run(job.disconnect())
        config.proximl.run(job.refresh())
        click.echo(f"Endpoint is running at:  {job.url}", file=config.stdout)
    else:
        if connect:
            click.echo("Waiting for job to start...", file=config.stdout)
            config.proximl.run(job.wait_for("running", timeout))
            config.proximl.run(job.refresh())
            click.echo(
                f"Endpoint is running at:  {job.url}", file=config.stdout
            )



import json
import logging


class Devices(object):
    def __init__(self, proximl):
        self.proximl = proximl

    async def get(self, provider_uuid, region_uuid, id, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device/{id}",
            "GET",
            kwargs,
        )
        return Device(self.proximl, **resp)

    async def list(self, provider_uuid, region_uuid, **kwargs):
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device",
            "GET",
            kwargs,
        )
        devices = [Device(self.proximl, **device) for device in resp]
        return devices

    async def create(
        self,
        provider_uuid,
        region_uuid,
        friendly_name,
        hostname,
        minion_id,
        **kwargs,
    ):
        logging.info(f"Creating Device {friendly_name}")
        data = dict(
            friendly_name=friendly_name,
            hostname=hostname,
            minion_id=minion_id,
            type="device",
            service="compute",
            **kwargs,
        )
        payload = {k: v for k, v in data.items() if v is not None}
        resp = await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device",
            "POST",
            None,
            payload,
        )
        device = Device(self.proximl, **resp)
        logging.info(f"Created Device {friendly_name} with id {device.id}")
        return device

    async def remove(self, provider_uuid, region_uuid, id, **kwargs):
        await self.proximl._query(
            f"/provider/{provider_uuid}/region/{region_uuid}/device/{id}",
            "DELETE",
            kwargs,
        )


class Device:
    def __init__(self, proximl, **kwargs):
        self.proximl = proximl
        self._device = kwargs
        self._id = self._device.get("device_id")
        self._provider_uuid = self._device.get("provider_uuid")
        self._region_uuid = self._device.get("region_uuid")
        self._name = self._device.get("friendly_name")
        self._hostname = self._device.get("hostname")
        self._status = self._device.get("status")
        self._online = self._device.get("online")
        self._maintenance_mode = self._device.get("maintenance_mode")
        self._device_config_id = self._device.get("device_config_id")
        self._job_status = self._device.get("job_status")
        self._job_last_deployed = self._device.get("job_last_deployed")
        self._job_config_id = self._device.get("job_config_id")
        self._job_config_revision = self._device.get("job_config_revision")

    @property
    def id(self) -> str:
        return self._id

    @property
    def provider_uuid(self) -> str:
        return self._provider_uuid

    @property
    def region_uuid(self) -> str:
        return self._region_uuid

    @property
    def name(self) -> str:
        return self._name

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def status(self) -> str:
        return self._status

    @property
    def online(self) -> bool:
        return self._online

    @property
    def maintenance_mode(self) -> bool:
        return self._maintenance_mode

    @property
    def device_config_id(self) -> str:
        return self._device_config_id

    @property
    def job_status(self) -> str:
        return self._job_status

    @property
    def job_last_deployed(self) -> str:
        return self._job_last_deployed

    @property
    def job_config_id(self) -> str:
        return self._job_config_id

    @property
    def job_config_revision(self) -> str:
        return self._job_config_revision

    def __str__(self):
        return json.dumps({k: v for k, v in self._device.items()})

    def __repr__(self):
        return f"Device( proximl , **{self._device.__repr__()})"

    def __bool__(self):
        return bool(self._id)

    async def remove(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/{self._id}",
            "DELETE",
        )

    async def refresh(self):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/{self._id}",
            "GET",
        )
        self.__init__(self.proximl, **resp)
        return self

    async def toggle_maintenance(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/{self._id}/maintenance",
            "PATCH",
        )

    async def run_action(self, command):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/{self._id}/action",
            "POST",
            None,
            dict(command=command),
        )

    async def set_config(self, device_config_id):
        resp = await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/{self._id}",
            "PATCH",
            None,
            dict(device_config_id=device_config_id),
        )
        self.__init__(self.proximl, **resp)
        return self

    async def deploy_endpoint(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/{self._id}/deploy",
            "PUT",
        )

    async def stop_endpoint(self):
        await self.proximl._query(
            f"/provider/{self._provider_uuid}/region/{self._region_uuid}/device/{self._id}/stop",
            "PUT",
        )
